﻿


using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using CppDepend.CodeModel;
using CppDepend.Helpers;

namespace CppDepend.PowerTools.DeadCode {
   static class DeadCodeAlgo {
      internal static ICodeMetric<IType, ushort> FindDeadTypes(ICodeBase codeBase) {
         Debug.Assert(codeBase != null);
         // Filter function for types that should'nt be considered as dead
         var canTypeBeConsideredAsDead = new Func<IType, bool>(
            t => !t.IsPublic );

         // Select types unused
         var typesUnused =
            from t in codeBase.Application.Types where
            t.NbTypesUsingMe == 0 && canTypeBeConsideredAsDead(t)
            select t;

         // Dead types = types used only by unused types (recursive)
         var deadTypesMetric = typesUnused.FillIterative(
         types => from t in codeBase.Application.Types.UsedByAny(types).Except(types)
                  where canTypeBeConsideredAsDead(t) &&
                        t.TypesUsingMe.Intersect(types).Count() == t.NbTypesUsingMe
                  select t);
         return deadTypesMetric;
      }

      internal static ICodeMetric<IMethod, ushort> FindDeadMethods(ICodeBase codeBase) {
         Debug.Assert(codeBase != null);
       
         // Filter function for methods that should'nt be considered as dead
         var canMethodBeConsideredAsDead = new Func<IMethod, bool>(
             m => !m.IsPublic &&                // Public methods might be used by client applications of your projects.
                  !m.IsEntryPoint &&            // Main() method is not used by-design.
                  !m.IsClassConstructor &&      
                  !m.IsVirtual &&               
                  !(m.IsConstructor &&          // Don't take account of protected ctor that might be call by a derived ctors.
                    m.IsProtected) 
              );

         // Get methods unused
         var methodsUnused =
            from m in codeBase.Application.Methods where
            m.NbMethodsCallingMe == 0 &&
            canMethodBeConsideredAsDead(m)
            select m;

         // Dead methods = methods used only by unused methods (recursive)
         var deadMethodsMetric = methodsUnused.FillIterative(
            methods => {
               // Use a hashet to make Intersect calls much faster!
               var hashset = methods.ToHashSet();
               return from m in codeBase.Application.Methods.UsedByAny(methods).Except(methods)
                      where canMethodBeConsideredAsDead(m) &&
                         // Select methods called only by methods already considered as dead
                          hashset.Intersect(m.MethodsCallingMe).Count() == m.NbMethodsCallingMe
                      select m;
            }
         );
         return deadMethodsMetric;
      }

      internal static IEnumerable<IField> FindDeadFields(ICodeBase codeBase) {
         Debug.Assert(codeBase != null);
        
         return from f in codeBase.Application.Fields where
            f.NbMethodsUsingMe == 0 &&
            !f.IsPublic &&     // Although not recommended, public fields might be used by client applications of your projects.
            !f.IsEnumValue 
                select f;
      }
   }
}
